"""Link tables.

.. autosummary::
   :toctree: .

   BiosampleTechsample
   ProjectExperiment
   FileBiosample
   FileExperiment
   FileReadout

"""
from ._link import (  # noqa
    BiosampleTechsample,
    FileBiosample,
    FileExperiment,
    FileReadout,
    ProjectExperiment,
)



import base64
import hashlib
from pathlib import Path
from typing import Any, List, Optional, Set, Tuple, Union

import anndata as ad
import pandas as pd
from lamin_logger import logger
from lndb import settings as setup_settings
from lndb_storage import UPath, load_to_memory
from lndb_storage.object import infer_suffix, size_adata, write_to_file
from lnschema_core import Features
from lnschema_core import File as lns_File
from lnschema_core import Run, Storage

from ._settings import settings
from .dev._core import get_name_suffix_from_filepath
from .dev.db._add import add, get_storage_root_and_root_str
from .dev.db._select import select
from .schema._table import table_meta

NO_NAME_ERROR = """
Pass a name in `ln.File(..., name=name)` when ingesting in-memory data.
"""

NO_SOURCE_ERROR = """
Error: Please link a data source using the `source` argument.
Fix: Link a data source by passing a run, e.g., via

run = lns.Run(transform=transform)
file = ln.File(..., source=run)

Or, by calling ln.context.track(), which sets a global run context.

More details: https://lamin.ai/docs/faq/ingest
"""


def serialize(
    data: Union[Path, UPath, str, pd.DataFrame, ad.AnnData], name, format
) -> Tuple[Any, Union[Path, UPath], str, str]:
    """Serialize a data object that's provided as file or in memory."""
    # Convert str to either Path or CloudPath
    if isinstance(data, (str, Path, UPath)):
        filepath = UPath(data)  # returns Path for local
        if (
            isinstance(filepath, UPath)
            and setup_settings.instance.storage.root not in filepath.parents  # noqa
        ):
            raise ValueError(
                "Can only track objects in configured cloud storage locations."
                " Please call `lndb.set_storage('< bucket_name >')`."
            )
        memory_rep = None
        name, suffix = get_name_suffix_from_filepath(filepath)
    # For now, in-memory objects are always saved to local_filepath first
    # This behavior will change in the future
    elif isinstance(data, (pd.DataFrame, ad.AnnData)):
        if name is None:
            raise ValueError(NO_NAME_ERROR)
        memory_rep = data
        suffix = infer_suffix(data, format)
        # this is always local
        filepath = Path(f"{name}{suffix}")
        if suffix != ".zarr":
            write_to_file(data, filepath)
    else:
        raise NotImplementedError("Recording not yet implemented for this type.")
    return memory_rep, filepath, name, suffix


def get_hash(local_filepath, suffix, check_hash: bool = True):
    if suffix != ".zarr":  # if not streamed
        hash = hash_file(local_filepath)
        if not check_hash:
            return hash
        result = select(lns_File, hash=hash).all()
        if len(result) > 0:
            msg = f"A file with same hash is already in the DB: {result}"
            if settings.error_on_file_hash_exists:
                hint = (
                    "💡 You can make this error a warning:\n"
                    "    ln.settings.error_on_file_hash_exists = False"
                )
                raise RuntimeError(f"{msg}\n{hint}")
            else:
                logger.warning(msg)
    else:
        hash = None
    return hash


def get_features_records(
    parsing_id: str,
    features_ref: Any,
    df_curated: pd.DataFrame,
) -> List:
    import lnschema_bionty as bionty

    # insert species entry if not exists
    species = select(bionty.Species, name=features_ref.species).one_or_none()
    if species is None:
        species = add(bionty.Species(name=features_ref.species))

    model = table_meta.get_model(f"bionty.{features_ref.entity}")

    # all existing feature records of the species in the db
    stmt = (
        select(model)
        .where(getattr(model, parsing_id).in_(df_curated.index))
        .where(getattr(model, "species_id") == species.id)  # type:ignore
    )
    records = stmt.all()
    records_df = df_curated.index.intersection(stmt.df()[parsing_id])

    # new records to be appended
    new_ids = df_curated.index.difference(records_df)
    if len(new_ids) > 0:
        # mapped new_ids
        mapped = features_ref.df.loc[features_ref.df.index.intersection(new_ids)].copy()
        mapped.index.name = parsing_id
        if mapped.shape[0] > 0:
            for kwargs in mapped.reset_index().to_dict(orient="records"):
                kwargs["species_id"] = species.id  # type:ignore
                record = model(**kwargs)
                records.append(record)
        # unmapped new_ids
        unmapped = set(new_ids).difference(mapped.index)
        if len(unmapped) > 0:
            for i in unmapped:
                record = model(
                    **{parsing_id: i, "species_id": species.id}  # type:ignore
                )
                records.append(record)
    return records


def parse_features(df: pd.DataFrame, features_ref: Any, **curate_kwargs) -> None:
    """Link features to a knowledge table.

    Args:
        df: a DataFrame
        features_ref: Features reference class.
    """
    from bionty import CellMarker, Gene, Protein

    # Add and curate features against a knowledge table
    # column = None
    # if parsing_id in df.columns:
    #     column = parsing_id
    # else:
    #     logger.warning(f"{parsing_id} column not found, using index as features.")
    df_curated = features_ref.curate(df=df, **curate_kwargs)
    # TODO: fix in the next PR
    parsing_id = df_curated.index.name
    if parsing_id is None:
        if features_ref.entity == "gene":
            parsing_id = "ensembl_gene_id"
        elif features_ref.entity == "cell_marker":
            parsing_id = "name"

    # logging of curation
    n = df_curated["__curated__"].count()
    n_mapped = df_curated["__curated__"].sum()
    log = {  # noqa  TODO: store this somewhere in the db
        "feature": parsing_id,
        "n_mapped": n_mapped,
        "percent_mapped": round(n_mapped / n * 100, 1),
        "unmapped": df_curated.index[~df_curated["__curated__"]],
    }

    features_hash = hash_set(set(df_curated.index))

    features = select(
        Features,
        id=features_hash,
        type=features_ref.entity,
    ).one_or_none()
    if features is not None:
        return features  # features already exists!
    else:
        features = Features(id=features_hash, type=features_ref.entity)
        records = get_features_records(parsing_id, features_ref, df_curated)

        if isinstance(features_ref, Gene):
            for record in records:
                features.genes.append(record)
        elif isinstance(features_ref, Protein):
            for record in records:
                features.proteins.append(record)
        elif isinstance(features_ref, CellMarker):
            for record in records:
                features.cell_markers.append(record)

    return features


def get_features(file_privates, features_ref, **curate_kwargs):
    """Updates file in place."""
    memory_rep = file_privates["_memory_rep"]
    if memory_rep is None:
        memory_rep = load_to_memory(file_privates["_local_filepath"])
    try:
        df = getattr(memory_rep, "var")  # for AnnData objects
        if callable(df):
            df = memory_rep
    except AttributeError:
        df = memory_rep
    return parse_features(df, features_ref, **curate_kwargs)


def get_run(run: Optional[Run]) -> Run:
    if run is None:
        from ._context import context

        run = context.run
        if run is None:
            raise ValueError(NO_SOURCE_ERROR)
    # the following ensures that queried objects (within __init__)
    # behave like queried objects, only example right now: Run
    if hasattr(run, "_ln_identity_key") and run._ln_identity_key is not None:
        run._sa_instance_state.key = run._ln_identity_key
    return run


def get_path_size_hash(
    filepath: Union[Path, UPath],
    memory_rep: Optional[Union[pd.DataFrame, ad.AnnData]],
    suffix: str,
    check_hash: bool = True,
):
    cloudpath = None
    localpath = None

    path = UPath(filepath)  # returns Path for local

    if suffix == ".zarr":
        if memory_rep is not None:
            size = size_adata(memory_rep)
        else:
            if isinstance(path, UPath):
                cloudpath = filepath
                # todo: properly calculate size
                size = 0
            else:
                localpath = filepath
                size = sum(f.stat().st_size for f in path.rglob("*") if f.is_file())
        hash = None
    else:
        if isinstance(path, UPath):
            try:
                size = path.stat()["size"]
            # here trying to fix access issue with new s3 buckets
            except Exception as e:
                if path._url.scheme == "s3":
                    path = UPath(filepath, cache_regions=True)
                    size = path.stat()["size"]
                else:
                    raise e
            cloudpath = filepath
            hash = None
        else:
            size = path.stat().st_size
            localpath = filepath
            hash = get_hash(filepath, suffix, check_hash=check_hash)

    return localpath, cloudpath, size, hash


# expose to user via ln.File
def get_file_kwargs_from_data(
    data: Union[Path, UPath, str, pd.DataFrame, ad.AnnData],
    *,
    name: Optional[str] = None,
    source: Optional[Run] = None,
    format: Optional[str] = None,
    # backward compat
    features_ref: Optional[Any] = None,
    **curate_kwargs,
):
    run = get_run(source)
    memory_rep, filepath, name, suffix = serialize(data, name, format)
    localpath, cloudpath, size, hash = get_path_size_hash(filepath, memory_rep, suffix)

    # if local_filepath is already in the configured storage location
    # skip the upload
    _, root_str = get_storage_root_and_root_str()
    storage = select(Storage, root=root_str).one()

    file_privates = dict(
        _local_filepath=localpath,
        _cloud_filepath=cloudpath,
        _memory_rep=memory_rep,
    )

    # TODO: remove later
    # backward compat
    if features_ref is not None:
        logger.warning(
            "DeprecationWarning: `features_ref` is deprecated, please use"
            " `ln.Features`!"
        )
        features = [
            get_features(file_privates, features_ref, **curate_kwargs)
        ]  # has to be list!
    else:
        features = []

    file_kwargs = dict(
        name=name,
        suffix=suffix,
        hash=hash,
        source_id=run.id,
        size=size,
        storage_id=storage.id,
        source=run,
        features=features,
    )

    return file_kwargs, file_privates


# expose to user via ln.Features
def get_features_from_data(
    data: Union[Path, UPath, str, pd.DataFrame, ad.AnnData],
    reference: Any,
    format: Optional[str] = None,
):
    memory_rep, filepath, _, suffix = serialize(data, "features", format)
    localpath, cloudpath, _, _ = get_path_size_hash(
        filepath, memory_rep, suffix, check_hash=False
    )

    file_privates = dict(
        _local_filepath=localpath,
        _cloud_filepath=cloudpath,
        _memory_rep=memory_rep,
    )
    return get_features(file_privates, reference)


def to_b64_str(bstr: bytes):
    b64 = base64.urlsafe_b64encode(bstr).decode().strip("=")
    return b64


# a lot to read about this: lamin-notes/2022/hashing
def hash_set(s: Set[str]) -> str:
    bstr = ":".join(sorted(s)).encode("utf-8")
    # as we're truncating at 20 b64, we choose md5 over sha512
    return to_b64_str(hashlib.md5(bstr).digest())[:20]


def hash_file(path: Path) -> str:
    # based on https://stackoverflow.com/questions/3431825/generating-an-md5-hash-of-a-file  # noqa
    hash_md5 = hashlib.md5()
    with open(path, "rb") as file:
        for chunk in iter(lambda: file.read(4096), b""):
            hash_md5.update(chunk)
    return to_b64_str(hash_md5.digest())



from pathlib import Path
from typing import List, Optional, Union

import lnschema_core
import nbproject
from lamin_logger import logger
from lndb import settings
from lndb.dev import InstanceSettings
from lnschema_core import Run, Transform, dev
from nbproject._is_run_from_ipython import is_run_from_ipython


class context:
    """Global run context.

    If you set `ln.context.transform = transform`, the `transform` record is
    recognized whenever you create a run with `ln.Run()`.

    There are two convenience methods for creating `transform` records for
    Jupyter notebooks and pipelines, each.

    You can load the latest run if it doesn't exist via:
    `ln.Run(load_latest=True)`.
    """

    instance: Optional[InstanceSettings] = None
    """Current instance."""
    transform: Optional[Transform] = None
    """Current transform."""
    run: Optional[Run] = None
    """Current run."""

    @classmethod
    def _track(
        cls,
        *,
        id: Optional[str] = None,
        version: Optional[str] = None,
        name: Optional[str] = None,
        transform: Optional[Transform] = None,
        load_latest_run: bool = False,
        notebook_path: Optional[str] = None,
        pypackage: Union[str, List[str], None] = None,
        editor: Optional[str] = None,
    ) -> None:
        """Track `Transform` & `Run` records for a notebook or pipeline.

        Stores them as `ln.context.transform` and `ln.context.run`.

        Call without arguments in most settings.

        If a Jupyter notebook has no associated metadata, attempts to write
        metadata to disk.

        Args:
            id: Transform id.
            version: Transform version.
            name: Transform name.
            transform: Can be "pipeline" or "notebook".
            load_latest_run: If True, loads latest run of transform.
            pypackage: One or more python packages to track.
            notebook_path: Filepath of notebook. Only needed if inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        cls.instance = settings.instance
        logger.info(f"Instance: {cls.instance.identifier}")
        logger.info(f"User: {settings.user.handle}")
        import lamindb as ln

        if is_run_from_ipython and transform is None:
            cls._track_notebook(
                pypackage=pypackage,
                filepath=notebook_path,
                id=id,
                v=version,
                name=name,
                editor=editor,
            )
        elif transform is None:
            if name is not None:
                cls._track_pipeline(name=name, version=version)
            else:
                raise ValueError("Pass `name` to track pipeline!")
        else:
            transform_exists = ln.select(Transform, id=transform.id).one_or_none()
            if transform_exists is None:
                transform_exists = ln.add(transform)
                logger.info(f"Added transform: {transform}")
            else:
                logger.info(f"Loaded transform: {transform_exists}")
            cls.transform = transform_exists
        Run(load_latest=load_latest_run)

    @classmethod
    def _track_notebook(
        cls,
        *,
        id: Optional[str] = None,
        v: Optional[str] = "0",
        name: Optional[str] = None,
        filepath: Optional[str] = None,
        pypackage: Optional[Union[str, List[str]]] = None,
        editor: Optional[str] = None,
    ):
        """Infer Jupyter notebook metadata and create `Transform` record.

        Args:
            id: Pass a notebook id manually.
            v: Pass a notebook version manually.
            name: Pass a notebook name manually.
            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            editor: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
        """
        cls.instance = settings.instance
        # original location of this code was _nb
        # legacy code here, see duplicated version in _run
        if id is None and name is None:
            nbproject_failed_msg = (
                "Auto-retrieval of notebook name & title failed.\nPlease paste error"
                " at: https://github.com/laminlabs/nbproject/issues/new \n\nFix: Run"
                f" ln.track(id={dev.id.notebook()}, name='my-notebook-name')"
            )
            try:
                nbproject.header(
                    pypackage=pypackage, filepath=filepath, env=editor, display=False
                )
            except Exception:
                raise RuntimeError(nbproject_failed_msg)
            # this contains filepath if the header was run successfully
            from nbproject._header import _filepath

            id = nbproject.meta.store.id
            v = nbproject.meta.store.version
            name = Path(_filepath).stem
            title = nbproject.meta.live.title
        elif id is None or name is None:
            # Both id and name need to be passed if passing it manually
            raise RuntimeError("Fix: Pass both id & name to ln.nb.header().")
        else:
            title = None

        import lamindb as ln

        transform = ln.select(
            Transform,
            id=id,
            v=v,
        ).one_or_none()
        if transform is None:
            transform = Transform(
                id=id,
                v=v,
                name=name,
                title=title,
                type="notebook",
            )
            transform = ln.add(transform)
            logger.info(f"Added notebook: {transform}")
        else:
            logger.info(f"Loaded notebook: {transform}")
            if transform.name != name or transform.title != title:
                response = input(
                    "Updated notebook name and/or title: Do you want to assign a new id"
                    " or version? (y/n)"
                )
                if response == "y":
                    print("Notebook metadata will be re-initialized.")
                    new_id, new_v = None, None
                    response = input("Do you want to generate a new id? (y/n)")
                    if response == "y":
                        new_id = lnschema_core.dev.id.notebook()
                    response = input(
                        "Do you want to set a new version (e.g. '1.1')? Type 'n' for"
                        " 'no'. (version/n)"
                    )
                    if new_v != "n":
                        if new_v == "y":
                            response = input("Please type the version: ")
                        new_v = response
                    if new_id is not None or new_v is not None:
                        nbproject.meta.store.id = new_id
                        nbproject.meta.store.version = new_v
                        nbproject.meta.store.write()
                        # at this point, depending on the editor, the process
                        # might crash that is OK as upon re-running, the
                        # notebook will have new metadata and will be registered
                        # in the db in case the python process does not exit, we
                        # need a new Notebook record
                        transform = Transform(id=id, v=v, type="notebook")

                transform.name = name
                transform.title = title
                ln.add(transform)

        # at this point, we have a transform object
        cls.transform = transform

    @classmethod
    def _track_pipeline(
        cls,
        name: str,
        *,
        version: Optional[str] = None,
    ):
        """Load or create pipeline record within `Transform`.

        Args:
            name: Name as used in `Transform.name`.
            version: Pipeline version. If `None`, load latest (sort by `created_at`).
        """
        cls.instance = settings.instance
        import lamindb as ln

        if version is not None:
            transform = ln.select(Transform, name=name, v=version).one()
        else:
            transform = (
                ln.select(Transform, name=name)
                .order_by(Transform.created_at.desc())
                .first()
            )
            if transform is None:
                response = input(
                    f"Did not find any pipeline record with name '{name}'. Create a new"
                    " one? (y/n)"
                )
                if response == "y":
                    transform = Transform(name=name, type="pipeline")
        cls.transform = transform



"""Access the schema.

- The core entities are the basis for tracking any data and are available from
  `lns.<entity>`.
- Additional mounted schema modules provide domain-specific entities and are
  available via `lns.<module>.<entity>`.

Import this submodule as::

   import lamindb.schema as lns

Core entities
=============

Users, projects, storage locations:

.. autosummary::
   :toctree: .

   User
   Project
   Storage

See the source code `here <https://lamin.ai/docs/lnschema-core>`__.

Exemplary extensions
====================

Any LaminDB schema module that has been mounted to an instance can be accessed like the bionty, wetlab, bfx modules below:

- `bionty <https://lamin.ai/docs/lnschema-bionty/api>`__: Knowledge-managed biological entities.
- `wetlab <https://lamin.ai/docs/lnschema-wetlab/api>`__: Wetlab operations.


Helper tools
============

.. autosummary::
   :toctree: .

   view
   list_tables
   dev

"""
import importlib as _importlib

from lndb import settings as _settings
from lndb.dev._setup_schema import get_schema_module_name as _get_schema_module_name
from lnschema_core import Features, Project, Run, Storage
from lnschema_core import Transform as _Transform
from lnschema_core import User, dev
from packaging import version as _v

_check_v = {
    "bionty": "0.12.1",
    "wetlab": "0.15rc1",
}

from .. import _instance_setup


def _import_schema():
    for name in _settings.instance.schema:
        _module = _importlib.import_module(_get_schema_module_name(name))
        if name in _check_v:
            if _v.parse(_module.__version__) != _v.parse(_check_v[name]):
                raise RuntimeError(f"lamindb needs lnschema_{name}=={_check_v[name]}")
        globals()[_module._name] = _module


if _instance_setup:
    _import_schema()

from ._core import list_tables, view

list_entities = list_tables  # backward compat

Pipeline = _Transform
Notebook = _Transform



"""v0.7.1.

Revision ID: a171e861e473
Revises: e2507b7564f0
Create Date: 2022-10-19 18:33:50.266432

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a171e861e473"
down_revision = "e2507b7564f0"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("experiment", schema=None) as batch_op:
        batch_op.drop_index("ix_experiment_project_id")
        batch_op.drop_constraint("fk_experiment_project_id_id", type_="foreignkey")
        batch_op.drop_column("project_id")

    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.drop_index("ix_project_created_at")
        batch_op.drop_index("ix_project_created_by")
        batch_op.drop_index("ix_project_external_id")
        batch_op.drop_index("ix_project_name")
    op.drop_table("project")


def downgrade() -> None:
    pass



from pathlib import Path
from subprocess import run


def pytest_sessionstart(session):
    instance_dirs = [
        d for d in ["./docs/guide/mydata", "./mydata-test-db"] if Path(d).exists()
    ]
    for instance_dir in instance_dirs:
        clean_instance = f"rm -r {instance_dir}"
        session.run(*clean_instance.split(" "))
    run("lamin init --storage mydata-test-db", shell=True)



from lndb import settings
from lnschema_core.dev.sqlmodel import schema_sqlmodel
from sqlmodel import Field

from . import _name as schema_name

SQLModel, prefix, schema_arg = schema_sqlmodel(schema_name)


if "wetlab" in settings.instance.schema:

    class ProjectExperiment(SQLModel, table=True):  # type: ignore
        """Links `Project` and `Experiment`."""

        __tablename__ = f"{prefix}project_experiment"

        project_id: str = Field(foreign_key="core.project.id", primary_key=True)
        experiment_id: str = Field(foreign_key="wetlab.experiment.id", primary_key=True)

    class FileExperiment(SQLModel, table=True):  # type: ignore
        """Links for `File` and `Experiment`."""

        __tablename__ = f"{prefix}file_experiment"

        file_id: str = Field(foreign_key="core.file.id", primary_key=True)
        experiment_id: str = Field(foreign_key="wetlab.experiment.id", primary_key=True)

    class FileReadout(SQLModel, table=True):  # type: ignore
        """Links for `File` and `Readout`."""

        __tablename__ = f"{prefix}file_readout"

        file_id: str = Field(foreign_key="core.file.id", primary_key=True)
        readout_id: str = Field(foreign_key="wetlab.readout.id", primary_key=True)

    class BiosampleTechsample(SQLModel, table=True):  # type: ignore
        """Links for `Biosample` and `Techsample`."""

        __tablename__ = f"{prefix}biosample_techsample"

        biosample_id: str = Field(foreign_key="wetlab.biosample.id", primary_key=True)
        techsample_id: str = Field(foreign_key="wetlab.techsample.id", primary_key=True)

    class FileBiosample(SQLModel, table=True):  # type: ignore
        """Links for `File` and `Biosample`."""

        __tablename__ = f"{prefix}file_biosample"

        file_id: str = Field(foreign_key="core.file.id", primary_key=True)
        biosample_id: str = Field(foreign_key="wetlab.biosample.id", primary_key=True)

else:
    ProjectExperiment = None  # type: ignore
    FileExperiment = None  # type: ignore
    BiosampleTechsample = None  # type: ignore
    FileBiosample = None  # type: ignore
    FileReadout = None  # type: ignore



from typing import Optional

import bionty as bt
from lnschema_bionty import CellType, Disease, Species, Tissue
from lnschema_bionty.dev._bionty import knowledge
from lnschema_core import File
from lnschema_core.dev.sqlmodel import add_relationship_keys, schema_sqlmodel
from sqlalchemy.orm import relationship
from sqlmodel import Field, Relationship

from . import _name as schema_name
from ._core import (
    BiosampleBase,
    ExperimentBase,
    ExperimentTypeBase,
    ReadoutBase,
    TechsampleBase,
)
from ._link import BiosampleTechsample, FileBiosample, FileExperiment, FileReadout

_, prefix, schema_arg = schema_sqlmodel(schema_name)


class Experiment(ExperimentBase, table=True):  # type: ignore
    """Experiments."""

    experiment_type_id: str = Field(
        default=None, foreign_key="wetlab.experiment_type.id", index=True
    )
    files: File = Relationship(
        back_populates="experiments",
        sa_relationship_kwargs=dict(secondary=FileExperiment.__table__),
    )


File.experiments = relationship(
    Experiment, back_populates="files", secondary=FileExperiment.__table__
)
File.__sqlmodel_relationships__["experiments"] = None


class ExperimentType(ExperimentTypeBase, table=True):  # type: ignore
    """Experiment types."""

    __tablename__ = f"{prefix}experiment_type"


class Biosample(BiosampleBase, table=True):  # type: ignore
    """Biological samples that are registered in experiments."""

    batch: Optional[str] = None
    species_id: Optional[str] = Field(
        default=None, foreign_key="bionty.species.id", index=True
    )
    species: Species = Relationship()
    tissue_id: Optional[str] = Field(
        default=None, foreign_key="bionty.tissue.id", index=True
    )
    tissue: Tissue = Relationship()
    cell_type_id: Optional[str] = Field(
        default=None, foreign_key="bionty.cell_type.id", index=True
    )
    cell_type: CellType = Relationship()
    disease_id: Optional[str] = Field(
        default=None, foreign_key="bionty.disease.id", index=True
    )
    disease: Disease = Relationship()
    files: File = Relationship(
        back_populates="biosamples",
        sa_relationship_kwargs=dict(secondary=FileBiosample.__table__),
    )


File.biosamples = relationship(
    Biosample, back_populates="files", secondary=FileBiosample.__table__
)
File.__sqlmodel_relationships__["biosamples"] = None

add_relationship_keys(Biosample)


class Techsample(TechsampleBase, table=True):  # type: ignore
    biosamples: Biosample = Relationship(
        back_populates="techsamples",
        sa_relationship_kwargs=dict(secondary=BiosampleTechsample.__table__),
    )


Biosample.techsamples = relationship(
    Techsample, back_populates="biosamples", secondary=BiosampleTechsample.__table__
)
Biosample.__sqlmodel_relationships__["techsamples"] = None


@knowledge(bt.Readout)
class Readout(ReadoutBase, table=True):  # type: ignore
    """Biological samples that are registered in experiments."""

    files: File = Relationship(
        back_populates="readouts",
        sa_relationship_kwargs=dict(secondary=FileReadout.__table__),
    )


File.readouts = relationship(
    Readout,
    back_populates="files",
    secondary=FileReadout.__table__,
)
File.__sqlmodel_relationships__["readouts"] = None



from typing import Optional

from lamin_logger import logger
from lndb_storage import load_to_memory
from lnschema_core import File
from sqlalchemy.orm.session import object_session

from lamindb._context import context

from ._settings import settings
from .dev._core import filepath_from_file


# this is exposed to the user as File.load
def load(file: File, stream: bool = False, is_run_input: Optional[bool] = None):
    if stream and file.suffix not in (".h5ad", ".zarr"):
        logger.warning(f"Ignoring stream option for a {file.suffix} object.")
    if is_run_input is None:
        track_run_input = settings.track_run_inputs_upon_load
    else:
        track_run_input = is_run_input
    if track_run_input:
        if object_session(file) is None:
            raise ValueError("Need to load with session open to track as input.")
        if context.run is None:
            raise ValueError(
                "No global run context set. Call ln.context.track() or pass input run"
                " directly."
            )
        else:
            file.targets.append(context.run)
            session = object_session(file)
            session.add(file)
            session.commit()
    return load_to_memory(filepath_from_file(file), stream=stream)



"""v0.7.2.

Revision ID: a5d6b07f1401
Revises: a171e861e473
Create Date: 2022-10-20 10:48:15.014883

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a5d6b07f1401"
down_revision = "a171e861e473"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.create_table(
        "project_experiment",
        sa.Column("project_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.Column("experiment_id", sqlmodel.sql.sqltypes.AutoString(), nullable=False),
        sa.ForeignKeyConstraint(
            ["experiment_id"],
            ["experiment.id"],
            name=op.f("fk_project_experiment_experiment_id_experiment"),
        ),
        sa.ForeignKeyConstraint(
            ["project_id"],
            ["project.id"],
            name=op.f("fk_project_experiment_project_id_project"),
        ),
        sa.PrimaryKeyConstraint(
            "project_id", "experiment_id", name=op.f("pk_project_experiment")
        ),
    )

    with op.batch_alter_table("biometa", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False)
        )
        batch_op.create_index(
            batch_op.f("ix_biometa_created_by"), ["created_by"], unique=False
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_biometa_created_by_user"), "user", ["created_by"], ["id"]
        )

    with op.batch_alter_table("experiment", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False)
        )
        batch_op.create_index(
            batch_op.f("ix_experiment_created_by"), ["created_by"], unique=False
        )
        batch_op.create_foreign_key(
            batch_op.f("fk_experiment_created_by_user"), "user", ["created_by"], ["id"]
        )

    with op.batch_alter_table("experiment_type", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_experiment_type_name"), ["name"], unique=False
        )


def downgrade() -> None:
    pass



"""Data lineage (`yvzi`).

.. note::
   Please see the documented API reference `here <https://lamin.ai/docs/db/lamindb.schema>`__.

   This page just provides a list of entities.

Import the package::

   import lnschema_core

Entities:

.. autosummary::
   :toctree: .

   File
   Run
   Transform
   Folder
   User
   Storage
   Project
   Features

Development tools:

.. autosummary::
   :toctree: .

   dev
   link
"""
_schema_id = "yvzi"
_name = "core"
_migration = "5846a15d9241"
__version__ = "0.30rc3"

from . import dev, link
from ._core import Features, File, Folder, Project, Run, Storage, Transform, User

# backward compat
DObject = File
DFolder = Folder



from enum import Enum


class usage(str, Enum):
    """Data access types."""

    ingest = "ingest"
    insert = "insert"
    select = "select"
    update = "update"
    delete = "delete"
    load = "load"
    link = "link"


class TransformType(Enum):
    pipeline = "pipeline"
    notebook = "notebook"

    def __repr__(self):
        return self.name



from functools import partial
from pathlib import Path, PurePath
from typing import List, Optional, Tuple, Union, overload  # noqa

import sqlmodel as sqm
from lamin_logger import logger
from lndb import settings as setup_settings
from lndb_storage import UPath, store_object, write_adata_zarr
from lndb_storage._file import print_hook
from lnschema_core import File, Folder
from pydantic.fields import ModelPrivateAttr
from sqlalchemy.orm.attributes import set_attribute

from .._docs import doc_args
from ._core import file_to_sqm, get_session_from_kwargs
from ._select import select

add_docs = """
Insert or update data records.

Inserts a new :term:`record` if the corresponding row doesn't exist.
Updates the corresponding row with the record if it exists.

To update a row, query it with `.get` or `.select` and modify it before
passing it to `add`.

Guide: :doc:`/guide/add-delete`.

Example:

1) Add a record (by passing a record)
>>> ln.add(lns.Pipeline(name="My pipeline", v="1"))
Pipeline(id="0Cb86EZj", name="My pipeline", v="1", ...)

2) Update an existing record
>>> pipeline = ln.select(lns.Pipeline, id="0Cb86EZj").one()
>>> pipeline.name = "New name"
>>> ln.add(experiment)
Pipeline(id="0Cb86EZj", name="New name", v="1", ...)

3) Add a record if not exist in the DB
>>> # add a record if the metadata combination is not already exist in the DB
>>> # if exists, returns the existing record from the DB
>>> ln.add(lns.Pipeline, name="My pipeline", v="1")
>>> # under the hood, this runs a query first based on passed fields
>>> # equivalent to the following:
>>> pipeline = ln.select(lns.Pipeline, name="My pipeline", v="1").one_or_none()
>>> if pipeline is None:
>>>     ln.add(pipeline)

Args:
    record: One or multiple records as instances of `SQLModel`.
"""


@overload
def add(record: sqm.SQLModel) -> sqm.SQLModel:
    ...


# Currently seeing the following error without type ignore:
# Overloaded function signature 2 will never be matched: signature 1's parameter
# type(s) are the same or broader
@overload
def add(records: List[sqm.SQLModel]) -> List[sqm.SQLModel]:  # type: ignore
    ...


@overload
def add(  # type: ignore
    entity: sqm.SQLModel, **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    ...


@doc_args(add_docs)
def add(  # type: ignore
    record: Union[sqm.SQLModel, List[sqm.SQLModel]], **fields
) -> Union[sqm.SQLModel, List[sqm.SQLModel]]:
    """{}"""  # noqa
    session = get_session_from_kwargs(fields)
    if isinstance(record, list):
        records = record
    elif isinstance(record, sqm.SQLModel):
        records = [record]
    else:
        model = file_to_sqm(record)
        results = select(model, **fields).one_or_none()
        if results is None:
            records = [model(**fields)]
        else:
            logger.info(
                f"An existing {results.__class__.__name__} record is found in the DB:"
            )
            return results

    if session is None:  # assume global session
        session = setup_settings.instance.session()
        setup_settings.instance._cloud_sqlite_locker.lock()
        close = True
    else:
        close = False

    # commit metadata to database
    db_error = None
    for record in records:
        write_objectkey(record)
        # the following ensures that queried objects (within __init__)
        # behave like queried objects, only example right now: Run
        if hasattr(record, "_ln_identity_key") and record._ln_identity_key is not None:
            record._sa_instance_state.key = record._ln_identity_key
        session.add(record)
    try:
        session.commit()
    except Exception as e:
        db_error = e

    # upload data objects to storage
    added_records = []
    if db_error is None:
        added_records, upload_error = upload_committed_records(records, session)

    if close:
        session.close()
        setup_settings.instance._update_cloud_sqlite_file()
        setup_settings.instance._cloud_sqlite_locker.unlock()

    error = db_error or upload_error
    if error is not None:
        error_message = prepare_error_message(records, added_records, error)
        raise RuntimeError(error_message)
    elif len(added_records) > 1:
        return added_records
    else:
        return added_records[0]


def upload_committed_records(records, session):
    """Upload records in a list of database-commited records to storage.

    If any upload fails, subsequent records are cleaned up from the DB.
    """
    # make sure ALL records are up-to-date to enable accurate comparison
    # during metadata cleanup
    error = None
    added_records = []
    for record in records:
        session.refresh(record)

    # upload data objects
    for record in records:
        if isinstance(record, File) and hasattr(record, "_local_filepath"):
            try:
                upload_data_object(record)
            except Exception as e:
                error = e
                break
        added_records += [record]

    # clean up metadata for objects not uploaded to storage
    if error is not None:
        for record in records:
            if record not in added_records:
                session.delete(record)
        session.commit()

    return added_records, error


def prepare_error_message(records, added_records, error) -> str:
    if len(records) == 1 or len(added_records) == 0:
        error_message = (
            "An error occured. No entries were uploaded or committed"
            " to the database. See error message below.\n\n"
        )
    else:
        error_message = (
            "An error occured. The following entries have been"
            " successfully uploaded and committed to the database:\n"
        )
        for record in added_records:
            error_message += (
                f"- {', '.join(record.__repr__().split(', ')[:3]) + ', ...)'}\n"
            )
        error_message += "\nSee error message below.\n\n"
    error_message += f"{str(error)}"
    return error_message


def local_instance_storage_matches_local_parent(file: File):
    storage = setup_settings.instance.storage

    if file._local_filepath is not None:
        path = file._local_filepath
    else:
        path = file.path()

    if not isinstance(path, UPath):
        path = path.resolve()

    parents = [str(p) for p in path.parents]
    return str(storage.root) in parents


def get_storage_root_and_root_str(
    root: Optional[Union[Path, UPath]] = None
) -> Tuple[Union[Path, UPath], str]:
    if root is None:
        root = setup_settings.instance.storage.root
    root_str = root.as_posix()
    if isinstance(root, UPath):
        root_str = root_str.rstrip("/")
    return root, root_str


def filepath_to_relpath(
    root: Union[PurePath, Path], root_str: str, filepath: Union[Path, UPath]
) -> Union[PurePath, Path]:
    """Filepath to relative path of the root."""
    if isinstance(root, UPath):
        relpath = PurePath(filepath.as_posix().replace(root_str, ""))
    else:
        relpath = filepath.resolve().relative_to(root_str)

    return relpath


def filepath_to_objectkey(
    record: Union[File, Folder], filepath: Union[Path, UPath]
) -> str:
    root, root_str = get_storage_root_and_root_str()

    relpath = filepath_to_relpath(root=root, root_str=root_str, filepath=filepath)
    # for File, _objectkey is relative path to the storage root without suffix
    _objectkey = relpath.parent / record.name if isinstance(record, File) else relpath

    return _objectkey.as_posix()


def write_objectkey(record: sqm.SQLModel) -> None:
    """Write to _objectkey.

    An objectkey excludes the storage root and the file suffix.
    """

    def set_objectkey(record: Union[File, Folder], filepath: Union[Path, UPath]):
        _objectkey = filepath_to_objectkey(record=record, filepath=filepath)
        set_attribute(record, "_objectkey", _objectkey)

    # _local_filepath private attribute is only added
    # when creating File from data or Folder from folder
    if hasattr(record, "_local_filepath"):
        # for upsert
        if isinstance(record._local_filepath, ModelPrivateAttr):
            pass
        elif record._local_filepath is None:
            # cloud storage
            if record._cloud_filepath is not None:
                set_objectkey(record, record._cloud_filepath)
            # both _cloud_filepath and _local_filepath are None fir zarr
        # local storage
        else:
            # only set objectkey if it is configured
            if local_instance_storage_matches_local_parent(record):
                set_objectkey(record, record._local_filepath)


def upload_data_object(file) -> None:
    """Store and add file and its linked entries."""
    file_storage_key = f"{file.id}{file.suffix}"

    storage = setup_settings.instance.storage

    if file._local_filepath is not None:
        # - Look for _cloud_filepath, which is only not None if the passed filepath
        # was in the existing storage in the first place (errors within _record.py)
        # - Look for _local_filepath and check whether it's in existing storage before
        # trying to copy the file
        if (file._cloud_filepath is None) and (
            not local_instance_storage_matches_local_parent(file)
        ):
            store_object(file._local_filepath, file_storage_key)
    elif file.suffix == ".zarr" and file._memory_rep is not None:
        storagepath = storage.key_to_filepath(file_storage_key)
        print_progress = partial(print_hook, filepath=file.name)
        write_adata_zarr(file._memory_rep, storagepath, callback=print_progress)



"""Base classes for a wetlab schema (`vvhc`).

Import the package::

   import lnschema_wetlab

Base classes:

.. autosummary::
   :toctree: .

   ExperimentBase
   ExperimentTypeBase
   BiosampleBase
   TechsampleBase
   ReadoutBase

Examples of derived classes, typically configured:

.. autosummary::
   :toctree: .

   Experiment
   ExperimentTypeBase
   BiosampleBase
   TechsampleBase
   ReadoutBase

Development tools:

.. autosummary::
   :toctree: .

   dev
   link

"""
# This is lnschema-module vvhc.
_schema_id = "vvhc"
_name = "wetlab"
_migration = "bfda12fc80a8"
__version__ = "0.15rc1"

from . import dev, link
from ._core import (  # noqa
    Biosample,
    BiosampleBase,
    Experiment,
    ExperimentBase,
    ExperimentType,
    ExperimentTypeBase,
    Readout,
    ReadoutBase,
    Techsample,
    TechsampleBase,
)



from itertools import islice
from pathlib import Path, PurePath
from typing import List, Optional, Union

from lndb_storage import UPath
from lnschema_core import File as lns_File
from lnschema_core import Folder as lns_Folder
from lnschema_core import Run, Storage

from ._record import get_storage_root_and_root_str
from .dev._core import filepath_from_folder, get_name_suffix_from_filepath
from .dev.db._add import filepath_to_relpath, write_objectkey
from .dev.db._select import select


def get_folder_kwargs_from_data(
    folder: Union[Path, UPath, str],
    *,
    name: Optional[str] = None,
    source: Optional[Run] = None,
):
    folderpath = UPath(folder)
    cloudpath = folderpath if isinstance(folderpath, UPath) else None
    localpath = None if isinstance(folderpath, UPath) else folderpath
    folder_privates = dict(
        _local_filepath=localpath,
        _cloud_filepath=cloudpath,
    )

    # TODO: UPath doesn't list the first level files and dirs with "*"
    pattern = "" if isinstance(folderpath, UPath) else "*"

    files = []
    for f in folderpath.rglob(pattern):
        if f.is_file():
            dobj = lns_File(f, source=source)
            write_objectkey(dobj)
            files.append(dobj)

    folder_kwargs = dict(
        name=folderpath.name if name is None else name,
        files=files,
    )
    return folder_kwargs, folder_privates


# Exposed to users as Folder.tree()
def tree(
    dir_path: Union[Path, UPath, str],
    level: int = -1,
    limit_to_directories: bool = False,
    length_limit: int = 1000,
):
    """Given a directory Path object print a visual tree structure.

    Adapted from: https://stackoverflow.com/questions/9727673/list-directory-tree-structure-in-python  # noqa
    """
    space = "    "
    branch = "│   "
    tee = "├── "
    last = "└── "

    dir_path = UPath(dir_path)
    files = 0
    directories = 0

    def inner(dir_path: Path, prefix: str = "", level=-1):
        nonlocal files, directories
        if not level:
            return  # 0, stop iterating
        # this is needed so that the passed folder is not listed
        contents = [
            i
            for i in dir_path.iterdir()
            if i.as_posix().rstrip("/") != dir_path.as_posix().rstrip("/")
        ]
        if limit_to_directories:
            contents = [d for d in contents if d.is_dir()]
        pointers = [tee] * (len(contents) - 1) + [last]
        for pointer, path in zip(pointers, contents):
            if path.is_dir():
                yield prefix + pointer + path.name
                directories += 1
                extension = branch if pointer == tee else space
                yield from inner(path, prefix=prefix + extension, level=level - 1)
            elif not limit_to_directories:
                yield prefix + pointer + path.name
                files += 1

    print(dir_path.name)
    iterator = inner(dir_path, level=level)
    for line in islice(iterator, length_limit):
        print(line)
    if next(iterator, None):
        print(f"... length_limit, {length_limit}, reached, counted:")
    print(f"\n{directories} directories" + (f", {files} files" if files else ""))


# Exposed to users as Folder.get()
def get_file(
    folder: lns_Folder, relpath: Union[str, Path, List[Union[str, Path]]], **fields
):
    """Get files via relative path to folder."""
    if isinstance(relpath, List):
        relpaths = [PurePath(i) for i in relpath]
    else:
        abspath = relpath_to_abspath(folder=folder, relpath=PurePath(relpath))
        if abspath.is_dir():
            abspaths_files = list_files_from_dir(abspath)
            root, root_str = get_storage_root_and_root_str(filepath_from_folder(folder))
            relpaths = [
                filepath_to_relpath(root=root, root_str=root_str, filepath=i)
                for i in abspaths_files
            ]
        else:
            relpaths = [PurePath(relpath)]

    file_objectkeys = [relpath_to_objectkey(folder=folder, relpath=i) for i in relpaths]

    return select_by_objectkey(file_objectkeys=file_objectkeys, **fields)


def list_files_from_dir(dirpath: Union[Path, UPath]):
    """List all files recursively from a directory."""
    return [i for i in dirpath.rglob("*") if i.is_file()]


def relpath_to_objectkey(folder: lns_Folder, relpath: PurePath):
    """Convert a relative path of folder to an absolute path."""
    name, _ = get_name_suffix_from_filepath(relpath)
    objectkey = str(PurePath(folder._objectkey) / relpath.parent / name)
    return objectkey


def relpath_to_abspath(folder: lns_Folder, relpath: PurePath):
    return filepath_from_folder(folder) / relpath


def select_by_objectkey(file_objectkeys: List[str], **fields):
    # query for unique comb of (_filekey, storage, suffix)
    files = (
        select(lns_File, **fields)
        .where(lns_File._objectkey.in_(file_objectkeys))
        .join(Storage, root=get_storage_root_and_root_str()[1])
        .all()
    )
    # TODO: return files in the same order as the file_objectkeys
    return files



from pathlib import Path
from typing import Union

from cloudpathlib import CloudPath


def get_name_suffix_from_filepath(filepath: Union[Path, CloudPath]):
    suffix = "".join(filepath.suffixes)
    name = filepath.name.replace(suffix, "")
    return name, suffix


# add type annotations back asap when re-organizing the module
def storage_key_from_file(dobj):
    if dobj._objectkey is None:
        return f"{dobj.id}{dobj.suffix}"
    else:
        return f"{dobj._objectkey}{dobj.suffix}"


# add type annotations back asap when re-organizing the module
def filepath_from_file(dobj):
    from lndb import settings

    storage_key = storage_key_from_file(dobj)
    filepath = settings.instance.storage.key_to_filepath(storage_key)
    return filepath


# add type annotations back asap when re-organizing the module
def filepath_from_folder(folder):
    from lndb import settings

    return settings.instance.storage.key_to_filepath(folder._objectkey)



import base64
from pathlib import Path

from lamindb._record import hash_file, to_b64_str
from lamindb.dev._core import get_name_suffix_from_filepath


def test_get_name_suffix_from_filepath():
    # based on https://stackoverflow.com/questions/31890341/clean-way-to-get-the-true-stem-of-a-path-object  # noqa
    dataset = [
        ("a", "a", ""),
        ("a.txt", "a", ".txt"),
        ("archive.tar.gz", "archive", ".tar.gz"),
        ("directory/file", "file", ""),
        ("d.x.y.z/f.a.b.c", "f", ".a.b.c"),
        ("logs/date.log.txt", "date", ".txt"),
    ]
    for path, name, suffix in dataset:
        filepath = Path(path)
        assert name, suffix == get_name_suffix_from_filepath(filepath)


def test_compute_hash():
    dataset = [
        ("file_1.txt", "a", "DMF1ucDxtqgxw5niaXcmYQ"),
    ]
    for path, content, hash in dataset:
        filepath = Path(path)
        with open(path, "w") as file:
            file.write(content)
        assert hash == hash_file(filepath)
        filepath.unlink()


def test_base64():
    # the following can be commented out over time
    mytest = "test".encode()
    b64_str = to_b64_str(mytest)
    b64_str_padded = f"{b64_str}=="
    assert base64.urlsafe_b64decode(b64_str_padded.encode()).hex() == mytest.hex()



from datetime import datetime as datetime
from pathlib import Path
from typing import Any, List, Optional, Union, overload  # noqa

import anndata as ad
import pandas as pd
import sqlalchemy as sa
import sqlmodel
from cloudpathlib import CloudPath
from lamin_logger import logger
from pydantic.fields import PrivateAttr
from sqlmodel import Field, ForeignKeyConstraint, Relationship

from . import _name as schema_name
from ._link import FileFeatures, FolderFile, ProjectFolder, RunIn  # noqa
from ._timestamps import CreatedAt, UpdatedAt
from ._users import CreatedBy
from .dev import id as idg
from .dev._storage import filepath_from_file, filepath_from_folder
from .dev.sqlmodel import schema_sqlmodel
from .dev.type import TransformType

SQLModel, prefix, schema_arg = schema_sqlmodel(schema_name)


class User(SQLModel, table=True):  # type: ignore
    """User accounts.

    All data in this table is synched from the cloud user account to ensure a
    globally unique user identity.
    """

    id: str = Field(primary_key=True)
    """User ID. Typically 8-character base64."""  # noqa
    email: str = Field(index=True, unique=True)
    """Primary user email used for logging in."""
    handle: str = Field(nullable=False, index=True, unique=True)
    """User handle."""
    name: Optional[str] = Field(index=True)
    """Long display name."""
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class Storage(SQLModel, table=True):  # type: ignore
    """Storage locations.

    A file or run-associated file can be stored in any desired S3,
    GCP, Azure or local storage location. This table tracks these locations
    along with metadata.
    """

    id: str = Field(default_factory=idg.storage, primary_key=True)
    """Base62 char ID, generated by :func:`~lamindb.schema.dev.id.storage`."""
    root: str = Field(index=True)
    """Semantic identifier to the root of the storage location, like an s3 path, a local path, etc."""  # noqa
    type: Optional[str] = None
    """Local vs. s3 vs. gcp etc."""
    region: Optional[str] = None
    """Cloud storage region if applicable."""
    created_at: datetime = CreatedAt
    updated_at: Optional[datetime] = UpdatedAt


class Folder(SQLModel, table=True):  # type: ignore
    """Data folders, collections of data objects.

    In LaminDB, a data folder is a collection of data objects (`File`).
    """

    id: str = Field(default_factory=idg.folder, primary_key=True)
    name: str = Field(index=True)
    files: List["File"] = Relationship(  # type: ignore  # noqa
        back_populates="folders",
        sa_relationship_kwargs=dict(secondary=FolderFile.__table__),
    )
    """Collection of :class:`~lamindb.File`."""
    created_by: str = CreatedBy
    """Auto-populated link to :class:`~lamindb.schema.User`."""
    created_at: datetime = CreatedAt
    """Time of creation."""
    updated_at: Optional[datetime] = UpdatedAt
    """Time of last update."""

    # private attributes are needed here to prevent sqlalchemy error
    _local_filepath: Optional[Path] = PrivateAttr()
    _cloud_filepath: Optional[CloudPath] = PrivateAttr()

    def path(self) -> Union[Path, CloudPath]:
        """Path on storage."""
        return filepath_from_folder(self)

    def tree(
        self,
        level: int = -1,
        limit_to_directories: bool = False,
        length_limit: int = 1000,
    ) -> None:
        """Print a visual tree structure."""
        from lamindb._folder import tree

        return tree(
            dir_path=self.path(),
            level=level,
            limit_to_directories=limit_to_directories,
            length_limit=length_limit,
        )

    def get(self, relpath: Union[str, Path, List[Union[str, Path]]], **fields):
        """Get files via relative path to folder."""
        from lamindb._folder import get_file

        return get_file(folder=self, relpath=relpath, **fields)

    @overload
    def __init__(
        self,
        folder: Union[Path, str] = None,
        *,
        name: Optional[str] = None,
    ):
        """Initialize from folder."""
        ...

    @overload
    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        files: List["File"] = [],
    ):
        """Initialize from fields."""
        ...

    def __init__(  # type: ignore
        self,
        folder: Union[Path, str] = None,
        *,
        # continue with fields
        id: Optional[str] = None,
        name: Optional[str] = None,
        x: List["File"] = [],
    ):
        if folder is not None:
            from lamindb._folder import get_folder_kwargs_from_data

            kwargs, privates = get_folder_kwargs_from_data(
                folder=folder,
                name=name,
            )
            if id is not None:
                kwargs["id"] = id
        else:
            kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        super().__init__(**kwargs)
        if folder is not None:
            self._local_filepath = privates["_local_filepath"]
            self._cloud_filepath = privates["_cloud_filepath"]


Folder._objectkey = sa.Column("_objectkey", sqlmodel.sql.sqltypes.AutoString(), index=True)


class Project(SQLModel, table=True):  # type: ignore
    """Projects."""

    id: str = Field(default_factory=idg.project, primary_key=True)
    name: str = Field(index=True)
    created_by: str = CreatedBy
    """Auto-populated link to :class:`~lamindb.schema.User`."""
    created_at: datetime = CreatedAt
    """Time of creation."""
    updated_at: Optional[datetime] = UpdatedAt
    """Time of last update."""


class File(SQLModel, table=True):  # type: ignore
    """See lamindb.schema for docstring."""

    id: str = Field(default_factory=idg.file, primary_key=True)
    """Base62 char ID, generated by :func:`~lamindb.schema.dev.id.file`."""
    name: Optional[str] = Field(index=True)
    """Semantic name or title. Defaults to `None`."""
    suffix: Optional[str] = Field(default=None, index=True)
    """Suffix to construct the storage key. Defaults to `None`.

    This is a file extension if the `file` is stored in a file format.
    It's `None` if the storage format doesn't have a canonical extension.
    """

    size: Optional[int] = Field(default=None, sa_column=sa.Column(sa.BigInteger(), index=True))
    """Size in bytes.

    Examples: 1KB is 1e3 bytes, 1MB is 1e6, 1GB is 1e9, 1TB is 1e12 etc.
    """
    hash: Optional[str] = Field(default=None, index=True)
    """Hash (md5)."""

    # We need the fully module-qualified path below, as there might be more
    # schema modules with an ORM called "Run"
    source: "lnschema_core._core.Run" = Relationship(back_populates="outputs")  # type: ignore  # noqa
    """Link to :class:`~lamindb.Run` that generated the `file`."""
    source_id: str = Field(foreign_key="core.run.id", index=True)
    """The source run id."""
    storage_id: str = Field(foreign_key="core.storage.id", index=True)
    """The id of :class:`~lamindb.schema.Storage` location that stores the `file`."""
    features: List["Features"] = Relationship(
        back_populates="files",
        sa_relationship_kwargs=dict(secondary=FileFeatures.__table__),
    )
    """Link to feature sets :class:`~lamindb.Features`"""
    folders: List[Folder] = Relationship(
        back_populates="files",
        sa_relationship_kwargs=dict(secondary=FolderFile.__table__),
    )
    """Collection of :class:`~lamindb.Folder` that contain this file."""
    targets: List["lnschema_core._core.Run"] = Relationship(  # type: ignore  # noqa
        back_populates="inputs",
        sa_relationship_kwargs=dict(secondary=RunIn.__table__),
    )
    "Runs that use this file as input."
    created_at: datetime = CreatedAt
    """Time of creation."""
    updated_at: Optional[datetime] = UpdatedAt
    """Time of last update."""

    # private attributes are needed here to prevent sqlalchemy error
    _local_filepath: Optional[Path] = PrivateAttr()
    _cloud_filepath: Optional[CloudPath] = PrivateAttr()
    _memory_rep: Any = PrivateAttr()

    def path(self) -> Union[Path, CloudPath]:
        """Path on storage."""
        return filepath_from_file(self)

    def load(self, stream: bool = False, is_run_input: bool = False):
        """Load data object.

        Returns in-memory representation if configured (say, an `AnnData` object
        for an `h5ad` file).

        Otherwise, returns a path to a locally cached on-disk object (say, a
        `.jpg` file).
        """
        from lamindb._load import load as lnload

        return lnload(file=self, stream=stream, is_run_input=is_run_input)

    @overload
    def __init__(
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        *,
        name: Optional[str] = None,
        features: List["Features"] = [],
        source: Optional["Run"] = None,
        id: Optional[str] = None,
        format: Optional[str] = None,
    ):
        """Initialize from data."""
        ...

    @overload
    def __init__(
        self,
        id: Optional[str] = None,
        name: Optional[str] = None,
        source: Optional["Run"] = None,
        suffix: Optional[str] = None,
        size: Optional[int] = None,
        hash: Optional[str] = None,
        source_id: Optional[str] = None,
        storage_id: Optional[str] = None,
        features: List["Features"] = [],
        targets: List["Run"] = [],
    ):
        """Initialize from fields."""
        ...

    def __init__(  # type: ignore
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        *,
        features: List["Features"] = [],
        source: Optional["Run"] = None,
        format: Optional[str] = None,
        # continue with fields
        id: Optional[str] = None,
        name: Optional[str] = None,
        suffix: Optional[str] = None,
        size: Optional[int] = None,
        hash: Optional[str] = None,
        source_id: Optional[str] = None,
        storage_id: Optional[str] = None,
        targets: List["Run"] = [],
        # backward compat
        features_ref: Optional[Any] = None,
    ):
        if data is not None:
            from lamindb._record import get_file_kwargs_from_data

            kwargs, privates = get_file_kwargs_from_data(
                data=data,
                name=name,
                source=source,
                format=format,
                features_ref=features_ref,
            )
            if id is not None:
                kwargs["id"] = id
        else:
            kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        super().__init__(**kwargs)
        if data is not None:
            self._local_filepath = privates["_local_filepath"]
            self._cloud_filepath = privates["_cloud_filepath"]
            self._memory_rep = privates["_memory_rep"]
            # when features are passed with data
            if not isinstance(features, List):
                features = [features]
            self.features += features


File._objectkey = sa.Column("_objectkey", sqlmodel.sql.sqltypes.AutoString(), index=True)
File.__table__.append_constraint(sa.UniqueConstraint("storage_id", "_objectkey", "suffix", name="uq_storage__objectkey_suffix"))


class Transform(SQLModel, table=True):  # type: ignore
    """Data transformations.

    Jupyter notebooks.

    Jupyter notebooks (`notebooks`) represent one type of data transformation
    (`run`) and have a unique correspondence in `run`.

    IDs for Jupyter notebooks are generated through nbproject.

    Pipelines.

    A pipeline is typically versioned software that can perform a data
    transformation/processing workflow. This can be anything from typical
    workflow tools (Nextflow, Snakemake, Prefect, Apache Airflow, etc.) to
    simple (versioned) scripts.
    """

    id: str = Field(default_factory=idg.pipeline, primary_key=True)
    """Id."""  # noqa
    v: str = Field(default="1", primary_key=True)
    """Version identifier, defaults to `"1"`.

    Use this to label different versions of the same transform.

    Consider using `semantic versioning <https://semver.org>`__
    with `Python versioning <https://peps.python.org/pep-0440/>`__.
    """
    name: str = Field(index=True)
    """File name of the transform.
    """
    type: TransformType = Field(index=True, default=TransformType.pipeline)
    """File name of the transform.
    """
    title: Optional[str] = Field(index=True)
    """Title of the transform as generated by `nbproject.meta.title
    <https://lamin.ai/docs/nbproject/nbproject.dev.metalive#nbproject.dev.MetaLive.title>`__.
    """
    created_by: str = CreatedBy
    """Auto-populated link to :class:`~lamindb.schema.User`."""
    created_at: datetime = CreatedAt
    """Time of creation."""
    updated_at: Optional[datetime] = UpdatedAt
    """Time of last update."""


class Run(SQLModel, table=True):  # type: ignore
    """Runs of data transforms.

    A `run` is any transform of a `file`.

    Args:
        id: Optional[str] = None
        name: Optional[str] = None
        load_latest: bool = False - Load latest run for given notebook or pipeline.
        transform: Optional[Transform] = None
        inputs: List[File] = None
        outputs: List[File] = None

    It typically has inputs and outputs:

    - References to outputs are stored in the `file` table in the
      `source_id` column as a foreign key the `run`
      table. This is possible as every given `file` has a unique data source:
      the `run` that produced the `file`. However, note that a given
      `run` may output several `files`.
    - References to inputs are stored in the `run_in` table, a
      many-to-many link table between the `file` and `run` tables. Any
      `file` might serve as an input for many `runs`. Similarly, any
      `run` might have many `files` as inputs.
    """

    __table_args__ = (
        ForeignKeyConstraint(
            ["transform_id", "transform_v"],
            ["core.transform.id", "core.transform.v"],
        ),
        {"schema": schema_arg},
    )
    id: Optional[str] = Field(default_factory=idg.run, primary_key=True)
    """Base62 char ID & primary key, generated through :func:`~lamindb.schema.dev.id.run`."""  # noqa
    name: Optional[str] = Field(default=None, index=True)
    external_id: Optional[str] = Field(default=None, index=True)
    transform_id: Optional[str] = Field(default=None, index=True)
    transform_v: Optional[str] = Field(default=None, index=True)
    transform: Transform = Relationship()
    """Link to :class:`~lamindb.schema.Notebook`."""
    outputs: List["File"] = Relationship(back_populates="source")
    """Output data :class:`~lamindb.File`."""
    inputs: List["File"] = Relationship(back_populates="targets", sa_relationship_kwargs=dict(secondary=RunIn.__table__))
    """Input data :class:`~lamindb.File`."""
    created_by: str = CreatedBy
    """Auto-populated link to :class:`~lamindb.schema.User`."""
    created_at: datetime = CreatedAt
    """Time of creation."""

    _ln_identity_key: Optional[str] = PrivateAttr(default=None)
    # simulate query result

    def __init__(  # type: ignore
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        load_latest: bool = False,
        external_id: Optional[str] = None,
        transform: Optional[Transform] = None,
        inputs: List[File] = None,
        outputs: List[File] = None,
    ):
        kwargs = {k: v for k, v in locals().items() if v and k != "self"}

        import lamindb as ln

        global_context = False
        if transform is None:
            if ln.context.transform is not None:
                global_context = True
                transform = ln.context.transform
            else:
                raise ValueError("Either call `ln.Run(transform=transform)` or `ln.track(transform=...)`.")

        if not isinstance(transform, Transform):
            raise TypeError("transform needs to be of type Transform")

        run = None
        if load_latest:
            run = ln.select(ln.Run, transform_id=transform.id, transform_v=transform.v).order_by(ln.Run.created_at.desc()).first()
            if run is not None:
                logger.info(f"Loaded run: {run}")
        elif id is not None:
            run = ln.select(ln.Run, id=id).one_or_none()
            if run is None:
                raise NotImplementedError("You can currently only pass existing ids")

        if run is None:
            kwargs.update(dict(transform_id=transform.id, transform_v=transform.v))
            super().__init__(**kwargs)
            self._ln_identity_key = None
        else:
            super().__init__(**run.dict())
            self._ln_identity_key = run.id  # simulate query result

        if global_context:
            if run is None:
                added_self = ln.add(self)
                self._ln_identity_key = added_self.id
                logger.info(f"Added run: {self}")
            ln.context.run = self


class Features(SQLModel, table=True):  # type: ignore
    """Sets of features."""

    id: str = Field(primary_key=True)  # use a hash
    type: str  # was called entity_type
    created_by: str = CreatedBy
    created_at: datetime = CreatedAt
    files: List["File"] = Relationship(
        back_populates="features",
        sa_relationship_kwargs=dict(secondary=FileFeatures.__table__),
    )

    @overload
    def __init__(
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
    ):
        """Initialize from data."""
        ...

    @overload
    def __init__(
        self,
        id: Optional[str] = None,
        type: Optional[str] = None,
        files: List["File"] = [],
    ):
        """Initialize from fields."""
        ...

    def __init__(  # type: ignore
        self,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
    ):
        kwargs = {k: v for k, v in locals().items() if v and k != "self"}
        super().__init__(**kwargs)

    def __new__(
        cls,
        data: Union[Path, str, pd.DataFrame, ad.AnnData] = None,
        reference: Any = None,
        *,
        id: str = None,
        type: Any = None,
        # continue with fields
        files: List["File"] = [],
    ):
        if data is not None:
            from lamindb._record import get_features_from_data

            features = get_features_from_data(
                data=data,
                reference=reference,
            )
        else:
            features = super().__new__(cls)
        return features



from typing import List, Optional, Union

import nbproject as _nb

# from lamin_logger import logger
from lnschema_core import Run, Transform

from ._context import context


# this whole class is deprecated, see lamindb.context instead!
class nb:
    """Manage Jupyter notebooks.

    - Guide: :doc:`guide/run`
    - FAQ: :doc:`/faq/nb`

    For more background, see `nbproject <https://lamin.ai/docs/nbproject>`__.
    """

    transform: Transform = None  # Notebook of this Python session
    run: Run = None  # run of this Python session

    @classmethod
    def header(
        cls,
        *,
        run: Optional[str] = None,
        pypackage: Union[str, List[str], None] = None,
        filepath: Optional[str] = None,
        env: Optional[str] = None,
        id: Optional[str] = None,
        v: Optional[str] = "0",
        name: Optional[str] = None,
    ) -> Run:
        """Track the notebook & display metadata.

        Call without arguments in most settings.

        If the notebook has no nbproject metadata, initializes & writes metadata
        to disk.

        Args:
            run: If `None`, loads the latest run of the notebook (or creates on if
                there isn't any).

                If `"new"`, instantiates a new `Run` record.

            pypackage: One or more python packages to track.
            filepath: Filepath of notebook. Only needed if automatic inference fails.
            env: Editor environment. Only needed if automatic inference fails.
                Pass `'lab'` for jupyter lab and `'notebook'` for jupyter notebook,
                this can help to identify the correct mechanism for interactivity
                when automatic inference fails.
            id: Pass a notebook id manually.
            v: Pass a notebook version manually.
            name: Pass a notebook name manually.
        """
        # logger.warning(
        #     "DeprecationWarning: Please replace ln.nb.header() with ln.track()."
        # )
        context._track_notebook(
            pypackage=pypackage, filepath=filepath, id=id, v=v, name=name, editor=env
        )
        cls.transform = context.transform
        if run == "new":
            run = Run()
        elif run is None:
            run = Run(load_latest=True)
        else:
            raise ValueError("Pass 'new' to ln.nb.header().")
        cls.run = run
        return run

    @classmethod
    def publish(cls, version: str = None, i_confirm_i_saved: bool = False):
        """Publish the notebook.

        Args:
            version: Notebook version to publish.
            i_confirm_i_saved: Only relevant outside Jupyter Lab & Notebook as a
                safeguard against losing the editor buffer content.
        """
        from nbproject._publish import finalize_publish, run_checks_for_publish

        import lamindb as ln

        result = run_checks_for_publish(
            calling_statement="publish(", i_confirm_i_saved=i_confirm_i_saved
        )
        if result != "checks-passed":
            return result
        finalize_publish(calling_statement="publish(", version=version)
        # update DB
        transform = ln.select(Transform, id=cls.transform.id, v=cls.transform.v).one()
        # update version
        transform.title = _nb.meta.live.title
        if version != cls.transform.v:
            transform_add = Transform(
                id=transform.id, v=version, name=transform.name, title=transform.title
            )
        else:
            transform_add = transform
        ln.add(transform_add)
        if version != cls.transform.v:
            cls.run.transform_v = version
            ln.add(cls.run)
            ln.delete(transform)



"""v0.6.0.

Revision ID: 3ed88d3699fd
Revises:
Create Date: 2022-10-14 23:48:42.242156

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "3ed88d3699fd"
down_revision = None
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("project", schema=None) as batch_op:
        batch_op.add_column(
            sa.Column("created_by", sqlmodel.sql.sqltypes.AutoString(), nullable=False)
        )
        batch_op.create_index(
            batch_op.f("ix_project_created_by"), ["created_by"], unique=False
        )
        batch_op.create_foreign_key(
            "fk_created_by_user_id", "user", ["created_by"], ["id"]
        )

    with op.batch_alter_table("version_vvhc", schema=None) as batch_op:
        batch_op.create_index(
            batch_op.f("ix_version_vvhc_user_id"), ["user_id"], unique=False
        )


def downgrade() -> None:
    pass



from lamin_logger import logger
from lndb import __version__ as lndb_v
from lnschema_core import __version__ as lnschema_core_v
from nbproject import __version__ as nbproject_v
from packaging import version

if version.parse(lndb_v) < version.parse("0.37.5rc1"):
    raise RuntimeError("Upgrade lndb! pip install lndb>=0.37.5rc1")

if version.parse(lnschema_core_v) != version.parse("0.30rc3"):
    raise RuntimeError("lamindb needs lnschema_core==0.30rc3")

if version.parse(nbproject_v) < version.parse("0.8.3"):
    raise RuntimeError("lamindb needs nbproject>=0.8.3")

# ensure that the lamin package is not installed
try:
    import lamin  # noqa

    logger.warning(
        "Please,\n"
        " - replace `import lamin` with `import lamindb.setup as lnsetup`\n"
        " - run `pip uninstall lamin`\n"
        "lamindb.setup now has all of the lamin functionality\n"
        "The lamindb API and lamin API will be integrated soon!\n"
        "The CLI remains as is!"
    )
except ImportError:
    pass



"""v0.7.0.post1.

Revision ID: e2507b7564f0
Revises: 01360b6492cc
Create Date: 2022-10-19 18:31:12.131732

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "e2507b7564f0"
down_revision = "01360b6492cc"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("experiment", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_experiment_project_id_id",
            "project",
            ["project_id"],
            ["id"],
        )


def downgrade() -> None:
    pass



"""v0.6.0.post1.

Revision ID: 05efb2ed6192
Revises: 3ed88d3699fd
Create Date: 2022-10-19 12:30:17.936854

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "05efb2ed6192"
down_revision = "3ed88d3699fd"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("biosample", schema=None) as batch_op:
        batch_op.create_foreign_key(
            "fk_biosample_cell_marker_id_id",
            "cell_marker",
            ["cell_marker_id"],
            ["id"],
        )


def downgrade() -> None:
    pass



from lnschema_core import File
from lnschema_core.dev import id


def test_id():
    assert len(id.file()) == 20
    assert len(id.user()) == 8
    assert len(id.secret()) == 40
    assert len(id.usage()) == 24
    assert len(id.instance()) == 12


def test_objectkey():
    assert "_objectkey" in File.__table__.columns.keys()



"""v0.7.0.

Revision ID: 01360b6492cc
Revises: 05efb2ed6192
Create Date: 2022-10-19 12:54:37.784842

"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

# revision identifiers, used by Alembic.
revision = "01360b6492cc"
down_revision = "05efb2ed6192"
branch_labels = None
depends_on = None


def upgrade() -> None:
    with op.batch_alter_table("biosample", schema=None) as batch_op:
        batch_op.drop_index("ix_biosample_cell_marker_id")
        batch_op.drop_constraint("fk_biosample_cell_marker_id_id", type_="foreignkey")
        batch_op.drop_column("cell_marker_id")

    with op.batch_alter_table("experiment", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            batch_op.f("uq_experiment_external_id"), ["external_id"]
        )

    with op.batch_alter_table("experiment_type", schema=None) as batch_op:
        batch_op.create_unique_constraint(
            batch_op.f("uq_experiment_type_efo_id"), ["efo_id"]
        )

    with op.batch_alter_table("featureset", schema=None) as batch_op:
        batch_op.create_unique_constraint(batch_op.f("uq_featureset_name"), ["name"])


def downgrade() -> None:
    pass



import os

import nox
from laminci import upload_docs_dir
from lndb.test.nox import (
    build_docs,
    login_testuser1,
    login_testuser2,
    run_pre_commit,
    run_pytest,
)

nox.options.reuse_existing_virtualenvs = True


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session(python=["3.7", "3.8", "3.9", "3.10", "3.11"])
def build(session):
    login_testuser1(session)
    login_testuser2(session)
    # run with pypi install on main
    if "GITHUB_EVENT_NAME" in os.environ:
        if os.environ["GITHUB_EVENT_NAME"] != "push":
            # run with submodule install on a PR
            session.install("./lnschema-core[dev,test]")
            session.install("./lnschema-wetlab[dev,test]")
            session.install("./lndb-storage[dev,test]")
    session.install(".[dev,test]")
    run_pytest(session)
    build_docs(session)
    login_testuser1(session)
    upload_docs_dir()



from pathlib import Path
from typing import Union

import fsspec

from ._upath_ext import UPath


def _infer_filesystem(path: Union[Path, UPath, str]):
    path_str = str(path)

    if isinstance(path, UPath):
        fs = path.fs
    else:
        protocol = fsspec.utils.get_protocol(path_str)
        if protocol == "s3":
            fs_kwargs = {"cache_regions": True}
        else:
            fs_kwargs = {}
        fs = fsspec.filesystem(protocol, **fs_kwargs)

    return fs, path_str



"""v0.8.0.

Revision ID: bfda12fc80a8
Revises: a5d6b07f1401
Create Date: 2022-11-03
"""
import sqlalchemy as sa  # noqa
import sqlmodel  # noqa
from alembic import op

revision = "bfda12fc80a8"
down_revision = "a5d6b07f1401"
branch_labels = None
depends_on = None


def upgrade() -> None:
    bind = op.get_bind()
    if bind.engine.name == "sqlite":
        op.rename_table(old_table_name="biometa", new_table_name="wetlab.biometa")
        op.rename_table(
            old_table_name="dobject_biometa", new_table_name="wetlab.dobject_biometa"
        )
        op.rename_table(old_table_name="biosample", new_table_name="wetlab.biosample")
        op.rename_table(old_table_name="techsample", new_table_name="wetlab.techsample")
        op.rename_table(
            old_table_name="biosample_techsample",
            new_table_name="wetlab.biosample_techsample",
        )
        op.rename_table(old_table_name="readout", new_table_name="wetlab.readout")
        op.rename_table(old_table_name="treatment", new_table_name="wetlab.treatment")
        op.rename_table(
            old_table_name="project_experiment",
            new_table_name="wetlab.project_experiment",
        )
        op.rename_table(old_table_name="experiment", new_table_name="wetlab.experiment")
        op.rename_table(
            old_table_name="experiment_type",
            new_table_name="wetlab.experiment_type",
        )
    else:
        op.execute("alter table public.biometa set schema wetlab")
        op.execute("alter table public.dobject_biometa set schema wetlab")
        op.execute("alter table public.biosample set schema wetlab")
        op.execute("alter table public.techsample set schema wetlab")
        op.execute("alter table public.biosample_techsample set schema wetlab")
        op.execute("alter table public.readout set schema wetlab")
        op.execute("alter table public.treatment set schema wetlab")
        op.execute("alter table public.project_experiment set schema wetlab")
        op.execute("alter table public.experiment set schema wetlab")
        op.execute("alter table public.experiment_type set schema wetlab")


def downgrade() -> None:
    pass



"""LaminDB: Manage R&D data & analyses.

Import the package::

   import lamindb as ln
   import lamindb.schema as lns

The central class of the API is `File`, a wrapper for files, on-disk (`zarr`, etc.)
and in-memory data objects (`DataFrame`, `AnnData`, etc.).

.. autosummary::
   :toctree: .

   File
   Folder

Track runs of data transformations:

.. autosummary::
   :toctree: .

   Run
   Transform
   track

Tracking data by features:

.. autosummary::
   :toctree: .

   Features

Query & manipulate data:

.. autosummary::
   :toctree: .

   select
   add
   delete

Manipulate data with open session:

.. autosummary::
   :toctree: .

   Session

View DB content:

.. autosummary::
   :toctree: .

   view

Schema - entities and their relations:

.. autosummary::
   :toctree: .

   schema

Setup:

.. autosummary::
   :toctree: .

   setup

Developer API:

.. autosummary::
   :toctree: .

   context
   settings
   dev
"""

__version__ = "0.35rc1"  # denote a release candidate for 0.1.0 with 0.1rc1

# prints warning of python versions
from lamin_logger import logger as _logger
from lamin_logger import py_version_warning as _py_version_warning

_py_version_warning("3.8", "3.10")

from lndb import settings as _setup_settings
from lndb._migrate import check_deploy_migration as _check_migrate
from lndb.dev._settings_store import (
    current_instance_settings_file as _current_settings_file,
)

from . import _check_versions  # executes checks during import

_instance_setup = False
if (
    not _current_settings_file().exists()
    or _setup_settings.instance.storage.root is None
):
    _logger.warning(
        "You haven't yet setup an instance using the CLI. Please call"
        " `lamindb.setup.init()` or `lamindb.setup.load()`."
    )
else:
    try:
        _check_migrate(
            usettings=_setup_settings.user, isettings=_setup_settings.instance
        )
        _instance_setup = True
    except Exception:
        _logger.warning(
            "Your current instance cannot be reached, init or load a connectable"
            " instance."
        )

from lnschema_core import Features, File, Folder, Run, Transform  # noqa

file_doc = """Data objects in storage & memory.

- Guide: :doc:`/guide/track`
- FAQ: :doc:`/faq/ingest`

A `File` is typically instantiated from data using the arguments below.
It can also be instantiated like any other SQLModel by passing all required
fields directly.

Args:
   data: Filepath or in-memory data.
   name: Name of the data object, required if an in-memory object is passed.
   source: The source of the data object (a :class:`~lamindb.Run`).
   id: The id of the file.
   format: Whether to use `h5ad` or `zarr` to store an `AnnData` object.

Data objects (`files`) represent atomic datasets in object storage:
jointly measured observations of variables (features).
They are generated by running code, instances of :class:`~lamindb.Run`.

A `file` may contain a single observation, for instance, a single image.

Data objects often have canonical on-disk and in-memory representations. If
choices among these representations are made, a one-to-one mapping can be
achieved, which means that any given `file` has a default in-memory and
on-disk representation.

LaminDB offers meaningful default choices. For instance,

- It defaults to pandas DataFrames for in-memory representation of tables
  and allows you to configure loading tables into polars DataFrames.
- It defaults to the `.parquet` format for tables, but allows you to
  configure `.csv` or `.ipc`.

Some datasets do not have a canonical in-memory representation, for
instance, `.fastq`, `.vcf`, or files describing QC of datasets.

.. note:: Examples for storage ⟷ memory correspondence:

   - Table: `.csv`, `.tsv`, `.parquet`, `.ipc` (`.feather`) ⟷ `pd.DataFrame`, `polars.DataFrame`
   - Annotated matrix: `.h5ad`, `.h5mu`, `.zarrad` ⟷ `anndata.AnnData`, `mudata.MuData`
   - Image: `.jpg`, `.png` ⟷ `np.ndarray`, ...
   - Tensor: zarr directory, TileDB store ⟷ zarr loader, TileDB loader
   - Fastq: `.fastq` ⟷ /
   - VCF: `.vcf` ⟷ /
   - QC: `.html` ⟷ /

"""
File.__doc__ = file_doc


from lndb_storage import subset
from lndb_storage.object import lazy

from . import dev  # noqa
from . import schema  # noqa
from . import setup  # noqa
from ._context import context  # noqa

track = context._track  # noqa
from ._delete import delete  # noqa
from ._nb import nb  # noqa
from ._settings import settings
from ._view import view  # noqa
from .dev.db import Session  # noqa
from .dev.db._add import add  # noqa
from .dev.db._select import select  # noqa
